#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  player.py
#  
#  Copyright 2019 Stefan Born <born@math.tu-berlin.de>
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#  
#  

from __future__ import division

import pygame
import random
import math
import numpy as np

from neural import *
from constants import *

class Player:
    
    stamina = 0
    survived_steps = 0
    alive = True
    x = 0
    y = 0 
    world = None # verweist auf die world-map 
    
    def mutate(self, initial=False):
        '''Diese Methode hat zwei Varianten:
        Ist der Parameter initial==True,
        so wird eine Anfangskonfiguration erzeugt.
        Ist der Parameter initial==False,
        so wird ein gegebenes 'Genom' mutiert.
        '''
        raise NotImplementedError
        
    def set_start_position(self,x,y):
        
        raise NotImplementedError
        
    def move(self):
        '''Das ist die Handlung des Wesens. Gegenwärtig 
        kann es sich nur bewegen.'''
        
        raise NotImplementedError
        

class RandomPlayer(Player):

    #Initialisiert den Spieler
    def __init__(self,x,y,speed,name):
        self.x=x #x-Pos
        self.y=y #y-Pos
        self.size = 1 #Größe
        self.sight = 10 #Sichtweite
        self.fat = 0    #Die Dicke der Fettschicht der Kreaturen
        self.speed = 1 #Geschwindigkeit
        self.inner_temprature = 0  #Die Innere Temperatur der Kreaturen
        self.skin_temprature = 0   #Die Hauttemperatur der Kreaturen
        self.status = True #Aktiv oder nicht
        self.chunkpos = [0,0] #Speichert die Position des Chunks
        self.stamina = START_STAMINA #Ausdauer
        self.angle = 0 #Laufwinkel      
        self.anglechange = 45  
        self.rect=pygame.Rect(int(self.x)-int(self.size),int(self.y)-int(self.size),self.size*2,self.size*2)#Rechteck in Pygame        
        self.name=name #ID        
        self.vx = 0#??
        self.vy = 0
        self.status = True
        self.alive = True
        self.viewpoint = 0
        self.survived_steps = 0
        self.sichtfeld = 0
        self.viewangle = 60
        self.v = False
        self.world = None
        self.past = [[],[],[]]
        self.model = make_model()
        self.dist = []
        self.minangle = 0
        self.minspeed = 0
        self.maxangle = 0
        self.maxspeed = 0
        self.fit = 0
        self.color = (27,27,27)
        self.collected_food = 0
        self.total_stamina = 0
        

    def set_start_position(self,x,y):
        self.x=x #x-Pos
        self.y=y
        self.rect=pygame.Rect(int(self.x)-int(self.size),int(self.y)-int(self.size),self.size*2,self.size*2)#Rechteck in Pygame  
    
    def mutate(self,initial=False):
        '''Diese Methode hat zwei Varianten:
        
        Ist der Parameter initial==True,
        so wird eine Anfangskonfiguration erzeugt.
        Ist der Parameter initial==False,
        so wird ein gegebenes 'Genom' mutiert.
        '''
        
        if initial:
            self.speed = random.randint(1,2)
            self.size  = random.randint(3,6)
            self.sight = random.randint(10,30)*self.size
            self.angle = random.randint(0,360)
            self.fat   = random.randint(0,max(int(self.size/2),1))
            self.memory_length = random.randint(0,5)
            
        else:
            self.speed = self.speed+random.choice(range(-9,10))*0.01
            self.sight = self.sight+random.choice(range(-9,10))*0.01*self.sight
            self.size  = self.size+random.choice(range(-9,10))*0.02
            self.fat   = self.fat+random.choice(range(-9,10))*0.01
            self.memory_length = self.memory_length+random.choice([-1,0,1])
            if self.fat <0:
                self.fat==0
            elif self.fat >self.size/2:  #Fettschicht kann nicht dicker sein als ein viertel der Größe
                self.fat=self.fat/2
            
        self.death=random.randint(1,MAX_DEATH)
        self.stamina = START_STAMINA+self.fat*self.size
        self.max_stamina = self.stamina*STAMINA_KOEFFIZIENT

    def looking_forward(self,objliste):
        self.v = False
        self.angle=self.angle%360 
        viewpoint_x =self.x
        viewpoint_y =self.y
        
        def distance(x1,y1,x2,y2):
            return ((x1-x2)**2+(y1-y2)**2)**(1/2)
        for obj in objliste:    
            dist = distance(obj.x,obj.y,self.x,self.y)
            
            if dist > self.sight*(2)**1/2:
                self.v = False
            else:
                points = []
                points.append([obj.x-obj.size,obj.y-obj.size])
                points.append([obj.x+obj.size,obj.y-obj.size])
                points.append([obj.x-obj.size,obj.y+obj.size])
                points.append([obj.x+obj.size,obj.y+obj.size])
                i = 0
                
                angles=[]
                for point in points:
                    if(self.sight**2>((viewpoint_x-point[0])**2+(viewpoint_y-point[1]**2))):
                        if(viewpoint_x==point[0]):
                            if(viewpoint_y>point[1]):
                                angle=270
                            else:
                                angle=90
                        else:
                            angle=math.atan((viewpoint_y-point[1])/(viewpoint_x-point[0]))/math.pi*180
                            #print(angle)
                            if(viewpoint_x>point[0]):
                                angle+=180
                            elif(angle<0):
                                angle+=360
                        #print("c")
                        angles.append(angle)
                y=(-(-((self.sight**2)-(viewpoint_x-points[0][0])**2)**(1/2)-viewpoint_y))
                y2=(-(((self.sight**2)-(viewpoint_x-points[0][0])**2)**(1/2)-viewpoint_y))
                #print(y)
                #print(y2)
                if((self.sight**2>(viewpoint_x-points[0][0])**2)and points[0][1]+2*obj.size>y and(points[0][1]<y)):
                    if(viewpoint_x==points[0][0]):
                        if(viewpoint_y>y):
                            angle=270
                        else:
                            angle=90
                    else:
                        angle=math.atan((viewpoint_y-y)/(viewpoint_x-points[0][0]))/math.pi*180
                        if(viewpoint_x>points[0][0]):
                            angle+=180
                        elif(angle<0):
                            angle+=360
                    #print(y,"a",angle)
                    angles.append(angle)
                if(self.sight**2>(viewpoint_x-points[0][0])**2 and points[0][1]+2*obj.size>y2 and(points[0][1]<y2)):
                    if(viewpoint_x==points[0][0]):
                        if(viewpoint_y>y2):
                            angle=270
                        else:
                            angle=90
                    else:
                        angle=math.atan((viewpoint_y-y2)/(viewpoint_x-points[0][0]))/math.pi*180
                        if(viewpoint_x>points[0][0]):
                            angle+=180
                        elif(angle<0):
                            angle+=360
                    #print(y2,"a",angle)
                    angles.append(angle)
                
                y=(-(-((self.sight**2)-(viewpoint_x-points[1][0])**2)**(1/2)-viewpoint_y))
                y2=(-(((self.sight**2)-(viewpoint_x-points[1][0])**2)**(1/2)-viewpoint_y))
                #print(y)
                #print(y2)
                if((self.sight**2>(viewpoint_x-points[1][0])**2)and points[1][1]+2*obj.size>y and(points[1][1]<y)):
                    if(viewpoint_x==points[0][0]):
                        if(viepoint_y>y):
                            angle=270
                        else:
                            angle=90
                    else:
                        angle=math.atan((viewpoint_y-y)/(viewpoint_x-points[1][0]))/math.pi*180
                        if(viewpoint_x>points[1][0]):
                            angle+=180
                        elif(angle<0):
                            angle+=360
                    #print(y,"b",angle)
                    angles.append(angle)
                if((self.sight**2>(viewpoint_x-points[1][0])**2)and points[0][1]+2*obj.size>y2 and(points[0][1]<y2)):
                    if(viewpoint_x==points[0][0]):
                        if(viewpoint_y>y2):
                            angle=270
                        else:
                            angle=90
                    else:
                        angle=math.atan((viewpoint_y-y)/(viewpoint_x-points[0][0]))/math.pi*180
                        if(viewpoint_x>points[0][0]):
                            angle+=180
                        elif(angle<0):
                            angle+=360
                    #print(y2,"b",angle)
                    angles.append(angle)
                x=(-(-(((self.sight**2)-(viewpoint_y-points[0][1])**2)**(1/2))-viewpoint_x))
                x2=(-(((self.sight**2)-(viewpoint_y-points[0][1])**2)**(1/2)-viewpoint_x))
                #print(x)
                #print(x2)
                if((self.sight**2>=(viewpoint_y-points[0][1])**2)and points[0][0]+(2*obj.size)>x and(points[0][0]<x)):
                    if(viewpoint_x==x):
                        angle=90
                    else:
                        angle=math.atan((viewpoint_y-points[0][1])/(viewpoint_x-x))/math.pi*180
                    if(viewpoint_x>x):
                        angle+=180
                    elif(angle<0):
                        angle+=360
                    #print(x,"c",angle)
                    angles.append(angle)
                if((self.sight**2>(viewpoint_y-points[0][1])**2)and points[0][0]+(2*obj.size)>x2 and(points[0][0]<x2)):
                    if(viewpoint_x==x2):
                        angle=90
                    else:
                        angle=math.atan((viewpoint_y-points[0][1])/(viewpoint_x-x2))/math.pi*180
                    if(viewpoint_x>x2):
                        angle+=180
                    elif(angle<0):
                        angle+=360
                    #print(x2,"c",angle)
                    angles.append(angle)
                x=(-(-((self.sight**2)-(viewpoint_y-points[2][1])**2)**(1/2)-viewpoint_x))
                x2=(-(((self.sight**2)-(viewpoint_y-points[2][1])**2)**(1/2)-viewpoint_x))
            
                if((self.sight**2>=(viewpoint_y-points[2][1])**2)and points[2][0]+(2*obj.size)>x and(points[2][0]<x)):
                    if(viewpoint_x==x):
                        if(viewpoint_y>points[2][1]):
                            angle=270
                        else:
                            angle=90
                    else:
                        angle=math.atan((viewpoint_y-points[2][1])/(viewpoint_x-x))/math.pi*180
                        if(viewpoint_x>x):
                            angle+=180
                        elif(angle<0):
                            angle+=360
                    #print(x,"d",angle)
                    angles.append(angle)
                if((self.sight**2>=(viewpoint_y-points[2][1])**2)and points[2][0]+2*obj.size>x2 and(points[2][0]<x2)):
                    if(viewpoint_x==x2):
                        if(viewpoint_y>points[2][1]):
                            angle=270
                        else:
                            angle=90
                    else:
                        angle=math.atan((viewpoint_y-points[2][1])/(viewpoint_x-x2))/math.pi*180
                        if(viewpoint_x>x):
                            angle+=180
                        elif(angle<0):
                            angle+=360
                    #print(x2,"d",angle)
                    angles.append(angle)

            if(len(angles)!=0):
                if(max(angles)-min(angles)>250):
                    for angle in angles:
                        if(angle>250):
                            angle=angle-360
                if((max(angles)>(self.angle+self.viewangle))):
                    if(min(angles)>(self.angle+self.viewangle)):
                        self.sichtfeld=0
                    elif(min(angles)<(self.angle-self.viewangle)):
                        self.sichtfeld=1
                    else:
                        self.sichtfeld=((self.angle+self.viewangle)-min(angles))/(self.viewangle*2)
                elif(min(angles)<(self.angle-self.viewangle)):
                    if(max(angles)<(self.angle-self.viewangle)):
                        self.sichtfeld=0                        
                    #print(self.angle-self.viewangle)
                    else:
                        self.sichtfeld=(max(angles)-(self.angle-self.viewangle))/(self.viewangle*2)
                else:
                    self.sichtfeld=((max(angles)-(min(angles)))/(self.viewangle*2))
            else:
                if((max(angles)-min(angles))>180):
                    for angle in angles:
                        if(angle>270):
                            angle=angle-360
               # print(points)
                self.sichtfeld=0
                #print(self.x,self.y)
                #print(angles)
                #self.sichtfeld=max(angles)-min(angles)
                #print(self.sichtfeld)       
            #return((max(angles)-min(angles)/viewangle)
            #print(max(angles)-min(angles)/(2*self.viewangle))
            #print(self.sichtfeld)



                
    def sightcheck(self,dist,alpha):
        self.objangle = alpha
        #print(dist)
        if dist<=self.sight:
            #print(self.angle%360,alpha)
            if self.angle-self.viewangle<0:
                if (self.angle%360-self.viewangle)+360 <= alpha or alpha <= self.angle+self.viewangle:
                    self.v = True
                    self.angle=alpha
                    print("2",self.v)
            elif self.angle-self.viewangle>360:
                if (self.angle%360-self.viewangle)-360 <= alpha or alpha <= self.angle+self.viewangle:
                    self.v = True
                    #print(self.v)
                    self.angle=alpha
            else:
                if self.angle%360-self.viewangle <= alpha <= self.angle+self.viewangle:
                    self.v = True
                    #print(self.v)
                    self.angle=alpha
        
        self.sichtfeld = 0.9


    #bewegt den Spieler
    def move(self):
        #Überprüfe, ob die Maximale Stamina überschritten wird und verringere die Stamina, sollte dem so sein
        if self.stamina>self.max_stamina:
            self.stamina=self.max_stamina
        
        angle_old = self.angle
        info = self.world.get_info(self)  ## Verwerte ich noch nicht, kann das

        if len(info)!=0:
            self.looking_forward(info)
            a = self.sichtfeld
        else:
            a=0
            
        if self.survived_steps>self.death:
                self.alive=False
                

        elif self.stamina >= self.speed+self.size**2+self.sight*0.25:

            inputlist = []
            inputlist.append(a)
            if len(self.past[0])>4:
                for i in self.past[0][-5:]:
                    inputlist.append(1/(i+1))
                for i in self.past[1][-5:]:
                    inputlist.append(1/(i+1))
                for i in self.past[2][-5:]:
                    inputlist.append(1/(i+1))
            else:
                for i in range(15):
                    inputlist.append(0)
            #print(len(inputlist))
            #print(inputlist)
            #print(len(inputlist))
            #print(inputlist)
            new_angle , new_speed = self.model.calc(inputlist)

            #print(new_angle)
            #print(new_speed)
            #print("-------")
            if self.minangle > new_angle:
                self.minangle = new_angle
            if self.minspeed > new_speed:
                self.minspeed = new_angle

            if self.maxangle < new_angle:
                self.maxangle = new_angle
            if self.maxspeed < new_speed:
                self.maxspeed = new_speed
            self.angle += new_angle
            self.angle = self.angle%360
            self.speed += new_speed
            self.speed = min(max(0, self.speed),3)
            self.past[0].append(self.angle/360)
            self.past[1].append(self.speed/10)
            self.past[2].append(a/300)

            
            self.x+=self.speed*math.cos(self.angle/180*math.pi)
            self.y+=self.speed*math.sin(self.angle/180*math.pi)
            self.rect=pygame.Rect(int(self.x)-int(self.size),int(self.y)-int(self.size),self.size*2,self.size*2)

            ## Es ist wichtig, eine Modellierungs zu machen, die nicht
            ## schon ein eindeutiges Evolutionsergebnis nahelegt.
            ## Schneller zu gehen, sollte mehr Energie kosten.
            ## Ein größeres Areal abzufressen sollte eine Energie 
            ## kosten, die proportional zum Areal ist.        
            self.skin_temprature=self.inner_temprature-self.fat

            #Der Energieverlust durch Temperatur ergibt sich aus dem Umfang, also dem freiligenden Teil des Körpers
            energy_loss_through_temprature=(self.skin_temprature-TEMPRATURE)*8*self.size  

            if energy_loss_through_temprature<0:
                energy_loss_through_temprature=0

            #Der Stamina-verlust an sich:

            self.stamina -=  MIN_STAMINA_LOSS_PER_STEP+0.5*(abs(self.speed)**2+self.size)
            self.survived_steps += 1 
            #self.stamina -= 1*(5+energy_loss_through_temprature+0.1*(self.speed+0.5*self.size**2+self.sight*0.2) )
            self.total_stamina += self.stamina
            #self.survived_steps += 1 
        else:
                self.fit = self.total_stamina #self.survived_steps 
                self.alive = False



def main(args):
    return 0

if __name__ == '__main__':
    import sys
    sys.exit(main(sys.argv))
